
import { JsonPointer } from './json-pointer';
import { SectionIds } from './section-ids';

export interface MarkdownBuilderOptions {
	numbered_headings?: boolean | {
		prefix?: string;
	};
}

export type Content = string | (() => string);

export type HeadingDepth = 1 | 2 | 3 | 4 | 5 | 6;

const max_section_depth = 5;

export interface TableOfContentsItem {
	id: string | JsonPointer;
	label: string;
	depth: HeadingDepth;
}

export interface AttrsOptions {
	id?: string | JsonPointer;
	classes?: string[];
	attrs?: Record<string, string | JsonPointer>;
}

export interface HeadingOptions extends AttrsOptions {
	/** If set to true, this heading will not show up in the table of contents */
	no_table_of_contents?: boolean;
}

export interface LinkOptions extends AttrsOptions {
	text?: string;
}

export interface CodeBlockOptions {
	lang?: string;
	caption?: string;
	extra_depth?: number;
}

export class MarkdownBuilder {
	private ids?: SectionIds;
	private contents: Content[] = [ ];
	private toc_items: TableOfContentsItem[] = [ ];
	private remaining_section_depth = max_section_depth;

	constructor(
		private readonly options: MarkdownBuilderOptions = { }
	) {
		if (options.numbered_headings) {
			if (typeof options.numbered_headings === 'object') {
				this.ids = new SectionIds(options.numbered_headings.prefix || 'section-');
			}
		}
	}

	public as_str() {
		return this.contents.map((content) => {
			return typeof content === 'function' ? content() : content;
		}).join('');
	}



	// =====

	public raw(raw: string) {
		this.contents.push(raw);
	}

	public text(text: string) {
		// todo: escape text
		this.contents.push(text);
	}



	// ===== Block =====

	public p(text: string) {
		return this.raw(text + '\n\n');
	}

	public h1(text: string, opts: HeadingOptions = { }) {
		return this.heading(1, text, opts);
	}

	public h2(text: string, opts: HeadingOptions = { }) {
		return this.heading(2, text, opts);
	}

	public h3(text: string, opts: HeadingOptions = { }) {
		return this.heading(3, text, opts);
	}

	public h4(text: string, opts: HeadingOptions = { }) {
		return this.heading(4, text, opts);
	}

	public h5(text: string, opts: HeadingOptions = { }) {
		return this.heading(5, text, opts);
	}

	public h6(text: string, opts: HeadingOptions = { }) {
		return this.heading(6, text, opts);
	}

	public heading(level: 1 | 2 | 3 | 4 | 5 | 6, text: string, { id, classes, no_table_of_contents }: HeadingOptions) {
		let label = text;

		if (! id && this.ids) {
			let sec: string;
			({ id, sec } = this.ids.next(level, text));
			label = `${sec}. ${label}`;
		}

		let rendered = `${'#'.repeat(level)} ${text}${this.attrs_md({ id, classes })}\n\n`;

		if (! no_table_of_contents) {
			this.toc_items.push({
				id,
				depth: level,
				label: text,
			});
		}

		this.raw(rendered);
	}

	public hr() {
		this.raw('---\n\n');
	}



	// ===== Open Blocks (closed separately) =====

	public section(opts: AttrsOptions = { }) {
		const depth = this.remaining_section_depth--;

		if (! depth) {
			this.remaining_section_depth++;
			throw new Error(`max section depth (${max_section_depth}) exceeded`);
		}

		const fence = '!'.repeat(depth + 5);
		const attrs = this.attrs_md(opts);
		
		this.raw(`${fence}${attrs}\n`);
		
		let closed = false;
		return () => {
			if (closed) {
				throw new Error('attempted to close section twice');
			}

			closed = true;
			this.raw(fence + '\n\n');
			this.remaining_section_depth++;
		};
	}

	public code_block(opts: CodeBlockOptions = { }) {
		const fence = '`'.repeat(3 + (opts.extra_depth || 0));
		const caption = opts.caption ? ' ' + JSON.stringify(opts.caption) : '';

		this.raw(`${fence}${opts.lang || 'plain'}${caption}\n`);
		
		let closed = false;
		return () => {
			if (closed) {
				throw new Error('attempted to close code block twice');
			}

			closed = true;
			this.raw(fence + '\n\n');
		};
	}



	// ===== Inline =====

	public code(text: string) {
		return `\`${text.replace(/`/g, '\\`')}\``;
	}

	public math(text: string) {
		return `$${text}$`;
	}

	public link(url: string, opts: LinkOptions = { }) {
		if (! opts.text) {
			return `<${url}>`;
		}

		const attrs = this.attrs_md(opts);

		if (! attrs) {
			return `[${opts.text}](${url})`;
		}

		// todo: more links
	}

	public em(text: string) {
		return `_${text.replace(/_/g, '\\_')}_`;
	}





	// ===== Complex =====

	public cite() {
		// todo: citations/footnotes
	}

	public table_of_contents() {
		return () => {
			return this.toc_items.map(({ id, depth, label }) => {
				return '  '.repeat(depth - 1) + `[${label}](#${id})`;
			}).join('\n') + '\n';
		};
	}

	public dl(opts: AttrsOptions = { }) {
		const attrs = this.attrs_html(opts);
		this.raw(`<dl${attrs}>\n`);

		let closed = false;
		return () => {
			if (closed) {
				throw new Error('attempted to close <dl> twice');
			}

			closed = true;
			this.raw('</dl>\n\n');
		};
	}

	public dt(text: string, opts: AttrsOptions = { }) {
		const attrs = this.attrs_html(opts);
		this.raw(`<dt${attrs}>${text}</dt>`);
	}

	public dd(text: string, opts: AttrsOptions = { }) {
		const attrs = this.attrs_html(opts);
		this.raw(`<dd${attrs}>${text}</dd>\n`);
	}



	// =====

	private attrs_md(raw_attrs: AttrsOptions) {
		const attrs: string[] = [ ];
		
		if (raw_attrs.id) {
			attrs.push('#' + md_str(raw_attrs.id));
		}

		if (raw_attrs.classes) {
			for (const classname of raw_attrs.classes) {
				attrs.push('.' + classname);
			}
		}

		if (raw_attrs.attrs) {
			for (const [name, value] of Object.entries(raw_attrs.attrs)) {
				attrs.push(`:${name}=${value}`);
			}
		}

		return attrs.length ? ` {${attrs.join(' ')}}` : '';
	}

	private attrs_html(raw_attrs: AttrsOptions) {
		const attrs: string[] = [ ];
		
		if (raw_attrs.id) {
			attrs.push(`id="${str(raw_attrs.id)}"`);
		}

		if (raw_attrs.classes) {
			attrs.push(`class="${raw_attrs.classes.join(' ')}"`);
		}

		if (raw_attrs.attrs) {
			for (const [name, value] of Object.entries(raw_attrs.attrs)) {
				attrs.push(`${name}="${value}"`);
			}
		}

		return attrs.length ? ` ${attrs.join(' ')}` : '';
	}
}

function md_str(str: string | JsonPointer) : string {
	if (str instanceof JsonPointer) {
		return str.as_md_str();
	}

	return str;
}

function str(str: string | JsonPointer) : string {
	if (str instanceof JsonPointer) {
		return str.as_str();
	}

	return str;
}
